print("import: 'intake'")
import intake

print("import: 'intake.auth'")
import intake.auth

print("import: 'intake.catalog'")
import intake.catalog

print("import: 'intake.source'")
import intake.source

